# -*- coding: utf-8 -*-

__author__ = """Larissa Triess"""
__email__ = 'mail@triess.eu'

import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg
from mpl_toolkits import axes_grid1


def confusion_matrix_image(
        tf_conf_mat: tf.Tensor, class_names: list, fig_dpi: int = 320, fig_size: int = 3,
        normalize: bool = True, cmap: plt.cm = plt.cm.Blues, scope: str = None
):

    """
    Visualization of the confusion matrix as an image to publish in TensorBoard with tf.summary.image

    Parameters:
        tf_conf_mat (tf.Tensor): A confusion matrix defined as
            Tensor(shape=[len(class_names), len(class_names)], dtype=tf.float32).
        class_names (list): The names of the classes (str).
        fig_dpi (int): The resolution of the output image.
        fig_size (int): The size of the figure.
        normalize (bool): If `True` numbers shown in the image will to normalized to 100%,
            otherwise absolute numbers will be depicted. Default is `True`.
        cmap (matplotlib.pyplot.cm): A `matplotlib` color map with which to paint the image.
            Default to `matplotlib.pyplot.cm.Blues`.
        scope (str): The name scope of the function. Defaults to `confusion_matrix_image`.

    Returns:
        Tensor(shape=[1, fig_dpi * fig_size, fig_dpi * fig_size, 3], dtype=tf.uint8): An RGB image.
            The first dimension corresponds to the `batch_size` to be published with `tf.summary.image`.

    Raises:
        ValueError: If length of class_names does not match the dimensions of tf_conf_mat.
    """

    if not tf_conf_mat.shape == [len(class_names)]*2:
        raise ValueError('Shape missmatch between `class_names` and `tf_conf_mat`. '
                         'The number of elements in `class_names` must be equal to the shape of '
                         'each dimension of `tf_conf_mat`.')

    with tf.name_scope(name=scope, default_name='confusion_matrix_image'):

        def add_colorbar(im, aspect=20, pad_fraction=2.0):
            divider = axes_grid1.make_axes_locatable(im.axes)
            width = axes_grid1.axes_size.AxesY(im.axes, aspect=0.6 / aspect)
            pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
            current_ax = plt.gca()
            cax = divider.append_axes("right", size=width, pad=pad)
            plt.sca(current_ax)
            return im.axes.figure.colorbar(im, cax=cax)

        def process_matrix(conf_mat):
            conf_mat = conf_mat.numpy()  # eager tensor to numpy array
            # normalize the confusion matrix (numbers will be displayed in percent)
            if normalize:
                _conf_mat = conf_mat
                conf_mat_sum = _conf_mat.sum(axis=1)
                _conf_mat[conf_mat_sum > 0] /= conf_mat_sum[conf_mat_sum > 0, None]
                conf_mat = _conf_mat * 100
                int_rest = conf_mat - conf_mat.astype(np.int32)
                sum_rest = np.round(int_rest.sum(axis=1)).astype(np.int32)
                rest_ordering = np.argsort(np.argsort(int_rest, axis=1), axis=1)
                rest_rounding = (
                    rest_ordering >= rest_ordering.shape[0] - sum_rest[:, None]
                )
                conf_mat = conf_mat.astype(np.int32) + rest_rounding.astype(np.int32)
                vminmax = dict(vmin=0, vmax=100)
            else:
                conf_mat = conf_mat.astype(np.float32)
                vminmax = dict()

            # plot confusion matrix to be displayed on TensorBoard
            np.set_printoptions(precision=2)  # limit display of floats to 2 decimals

            fig = Figure(figsize=(fig_size, fig_size), dpi=fig_dpi, facecolor='w', edgecolor='k')
            ax = fig.add_subplot(1, 1, 1)
            im = ax.imshow(conf_mat, interpolation='nearest', cmap=cmap, **vminmax)
            add_colorbar(im).ax.tick_params(labelsize=fig_size)

            ax.set_xlabel('Predicted Label', fontsize=fig_size)
            ax.set_xticks(np.arange(conf_mat.shape[1]))
            ax.set_xticklabels(class_names, fontsize=fig_size / 2 + 1, ha='right',
                               rotation=45, rotation_mode='anchor')
            ax.xaxis.set_label_position('top')
            ax.xaxis.tick_bottom()

            ax.set_ylabel('True Label', fontsize=fig_size)
            ax.set_yticks(np.arange(conf_mat.shape[0]))
            ax.set_ylim(conf_mat.shape[0] - 0.5, -0.5)
            ax.set_yticklabels(class_names, fontsize=fig_size / 2 + 1, va='center')
            ax.yaxis.set_label_position('right')
            ax.yaxis.tick_left()

            thresh = 2 * (conf_mat.max() / 3.0)
            for i in range(conf_mat.shape[0]):
                for j in range(conf_mat.shape[1]):
                    number_format = '{0:.1f}' if conf_mat.dtype == np.float32 else '{}'
                    ax.text(
                        j,
                        i,
                        number_format.format(conf_mat[i, j])
                        if conf_mat[i, j] != 0
                        else ".",
                        horizontalalignment="center",
                        fontsize=3 * fig_size / 4,
                        verticalalignment="center",
                        color="white" if conf_mat[i, j] > thresh else "black",
                    )

            canvas = FigureCanvasAgg(fig)
            canvas.draw()  # draw the canvas, cache the renderer
            image = np.fromstring(canvas.tostring_rgb(), dtype=np.uint8)
            image = image.reshape((fig_dpi * fig_size, fig_dpi * fig_size, 3))

            fig.clear()
            plt.close(fig)

            return image

        tf_image = tf.py_function(func=process_matrix,
                                  inp=[tf_conf_mat],
                                  Tout=tf.uint8,
                                  name='process_matrix')

        tf_image = tf.expand_dims(tf_image, 0)
        tf_image.set_shape([1, fig_dpi * fig_size, fig_dpi * fig_size, 3])

    return tf_image



# -*- coding: utf-8 -*-

__author__ = """Larissa Triess"""
__email__ = 'mail@triess.eu'

import math

import tensorflow as tf


def get_padding_sizes(
        spacial_dimensions: list or tuple, kernel_size: list or tuple, strides: list or tuple,
        scope: str = None
):

    """
    Computes the padding sizes for a rank 4 tensor according to the `kernel_size` and `strides`
    of a 2D convolution with `same` padding. The computation is equivalent to TF implementation.

    Parameters:
        spacial_dimensions (list | tuple): A tuple of two integers corresponding to the spacial
            dimensions H and W of Tensor(shape=[N, H, W, C]).
        kernel_size (list | tuple): A tuple of two integers defining the kernel size of a
            convolution for which to compute the corresponding padding of the input tensor.
        strides (list | tuple): A tuple of two integers defining the strides of a
            convolution for which to compute the corresponding padding of the input tensor.
        scope (str): The name scope of the function.

    Returns:
        (int, int, int, int): Four integers corresponding to the padding at the top, bottom,
            left, and right, respectively.

    Raises:
        TypeError: When `spacial_dimensions`, `kernel_size`, or `strides` do not have length 2.
        ValueError: If `strides` is smaller than 1 or greater than the spacial dimension.
        ValueError: If `kernel_size` is smaller than 1 or greater than the spacial dimension.
    """

    if not len(spacial_dimensions) == 2 or not all(isinstance(x, int) for x in spacial_dimensions):
        raise TypeError('`spacial_dimensions` must be a list or tuple of two integers.')
    if not len(kernel_size) == 2 or not all(isinstance(x, int) for x in kernel_size):
        raise TypeError('`kernel_size` must be a tuple of two integers.')
    if not len(strides) == 2 or not all(isinstance(x, int) for x in strides):
        raise TypeError('`strides` must be a tuple of two integers.')
    if not all([1 <= s <= d for s, d in zip(strides, spacial_dimensions)]):
        raise ValueError('Strides cannot be smaller than 1 or greater than the corresponding '
                         'spacial dimension. I got {}.'.format(strides))
    if not all([1 <= k <= d for k, d in zip(kernel_size, spacial_dimensions)]):
        raise ValueError('Kernel size cannot be smaller than 1 or greater than the corresponding '
                         'spacial dimension. I got {} vs. {}.'.format(kernel_size, spacial_dimensions))

    with tf.name_scope(name=scope, default_name='calculate_padding_sizes'):
        # Compute the output height and width
        out_h = int(math.ceil(float(spacial_dimensions[0]) / float(strides[0])))
        out_w = int(math.ceil(float(spacial_dimensions[1]) / float(strides[1])))

        # Calculate the amount of padding per spacial dimension
        pad_h = max((out_h - 1) * strides[0] + kernel_size[0] - spacial_dimensions[0], 0)
        pad_w = max((out_w - 1) * strides[1] + kernel_size[1] - spacial_dimensions[1], 0)

        pad_top = pad_h // 2  # amount of padding on the top
        pad_bottom = pad_h - pad_top  # amount of padding on the bottom
        pad_left = pad_w // 2  # amount of padding on the left
        pad_right = pad_w - pad_left  # amount of padding on the right

    return pad_top, pad_bottom, pad_left, pad_right


def pad(tensor, paddings, mode='CONSTANT', constant_values=0, name=None):
    """Pads a tensor.

    This operation pads a `tensor` according to the `paddings` you specify.
    `paddings` is a list of integers with shape `[n, 2]`, where n is the rank of
    `tensor`. For each dimension D of `input`, `paddings[D, 0]` indicates how
    many values to add before the contents of `tensor` in that dimension, and
    `paddings[D, 1]` indicates how many values to add after the contents of
    `tensor` in that dimension.

    The padded size of each dimension D of the output is:

    `paddings[D, 0] + tensor.dim_size(D) + paddings[D, 1]`

    For `mode` `CONSTANT`, `REFLECT`, and `SYMMETRIC` take a look at the TF documentation:
    https://www.tensorflow.org/api_docs/python/tf/pad

    If `mode` is "CYCLIC" then both `paddings[D, 0]` and `paddings[D, 1]` must be
    no greater than `tensor.dim_size(D)`.

    Example for `mode` `CYCLIC`:

    ```python
    t = tf.constant([[1, 2, 3], [4, 5, 6]])
    paddings = tf.constant([[0, 1], [2, 1]])
    tf.pad(t, paddings, "CYCLIC")  # [[2, 3, 1, 2, 3, 1],
                                   #  [5, 6, 4, 5, 6, 4],
                                   #  [2, 3, 1, 2, 3, 1]]
    ```

    Parameters:
        tensor: A `Tensor`.
        paddings: A list of integers.
        mode: One of "CONSTANT", "REFLECT", "SYMMETRIC", or "CYCLIC" (case-insensitive)
        name: A name for the operation (optional).
        constant_values: In "CONSTANT" mode, the scalar pad value to use. Must be
          same type as `tensor`.

      Returns:
        A `Tensor`. Has the same type as `tensor`.
    """
    tf_pads = tf.convert_to_tensor(paddings)

    if mode.upper() == 'CYCLIC':
        with tf.name_scope(name):
            rank = len(paddings)
            perm = [(i + 1) % rank for i in range(rank)]

            out = tensor
            for r in range(rank):
                # Create the padding slices and concat them to the original tensor
                out = tf.concat(
                    [out[tf.shape(out)[0] - tf_pads[r, 0]:, ...], out, out[:tf_pads[r, 1], ...]],
                    axis=0
                )

                # Move dimension of interest to position 0 for next step
                out = tf.transpose(out, perm)

            out.set_shape([None if t is None else t + p[0] + p[1] for t, p in
                           zip(tensor.get_shape().as_list(), paddings)])
    else:
        out = tf.pad(
            tensor,
            tf_pads,
            mode=mode,
            constant_values=constant_values,
            name=name
        )

    return out


def cyclic_padding(tensor: tf.Tensor, kernel_size: tuple, strides: tuple, scope: str = None):
    """
    This operation is a high-level wrapper for tf.pad(mode="CYCLIC") that computes the amount of
    padding for each dimension based on the `kernel_size` and `strides` specified. This function
    is intended to be a convenience wrapper when using convolutional layers with cyclic padding.
    It takes a tf.Tensor(shape=[N, H, W, C]) and performs zero padding in H and cyclic padding in W.

    For example:
    ```python
    t = tf.random.normal([4, 10, 50, 2])

    k_size = (3, 3)
    strides = (1, 1)

    t_padded = tf_helpers.cyclic_padding(t, k_size, strides)
    tf.layers.conv2d(t_padded, k_size, strides, padding='valid')

    # results in the same shape as
    # tf.layers.conv2d(t, k_size, strides, padding='same')
    ```

    Parameters:
        tensor: A `Tensor` of rank 4.
        kernel_size: A `tuple` of two integers.
        strides: A `tuple` of two integers.
        scope (str): A name for the operation.

    Returns:
        A `Tensor`. Has same type as `tensor`.
    """

    with tf.variable_scope(name_or_scope=scope, default_name='cyclic_padding'):

        inp_h, inp_w = tensor.get_shape().as_list()[1:3]

        pad_top, pad_bottom, pad_left, pad_right = get_padding_sizes(
            (inp_h, inp_w), kernel_size, strides)

        # Zero padding in the vertical dimension
        inp_pad_vertical = tf.pad(tensor, [[0, 0], [pad_top, pad_bottom], [0, 0], [0, 0]],
                                  mode='CONSTANT', constant_values=0)

        # Cyclic padding in the horizontal dimension
        out = pad(inp_pad_vertical,
                  [[0, 0], [0, 0], [pad_left, pad_right], [0, 0]],
                  mode='CYCLIC')
    return out



# -*- coding: utf-8 -*-

__author__ = """Larissa Triess"""
__email__ = 'mail@triess.eu'

import tensorflow as tf
from matplotlib.pylab import get_cmap


def expand_image_height(
        tf_image: tf.Tensor, factor: int, scope: str = None
):

    """
    Enlarges the input image-like tensor [H, W, C] of rank 3 in the first dimension (H)

    Parameters:
        tf_image (tf.Tensor): A rank 3 `Tensor` of shape [H, W, C].
        factor (int): The factor specifies how many multiples of each element in H are produced.
        scope (str): The name scope of the function.

    Returns:
        Tensor(shape=[H * factor, W, C], dtype=tf.image.dtype)

    Raises:
        NotImplementedError: When `tf_image` has rank 4. Batch dimension is not yet supported.
        ValueError: When `tf_image` has a different rank than 3 and 4.
        ValueError: If `factor` is not an integer greater or equal to 1.

    """

    shape = tf_image.shape

    if len(shape) == 4:
        raise NotImplementedError('Expand image height function does not support batched tensors.')
    if not len(shape) == 3:
        raise ValueError('The input tensor must be of rank 3.')
    if not isinstance(factor, int) or factor < 1:
        raise ValueError('Expected integer greater or equal to 1. I got {}.'.format(factor))

    with tf.name_scope(name=scope, default_name='expand_height'):
        shape = [shape[0] * factor, shape[1], shape[2]]
        tf_image = tf.expand_dims(tf_image, axis=1)
        tf_image = tf.tile(tf_image, multiples=[1, factor, 1, 1])
        tf_image = tf.reshape(tf_image, shape)
    return tf_image


def create_image(
        tensor: tf.Tensor, minval: float = 0.0, maxval: float = None,
        cmap: str = 'viridis', cycle_color_map: int = 1,
        mask: tf.Tensor = None, exclude_color: tuple = (0, 0, 0),
        expand_height: int = 1, scope: str = None,
):

    """
    Visualization of a two-dimensional tf.Tensor as a color image.

    Parameters:
        tensor (tf.Tensor): An image-like Tensor(shape=[H, W], dtype=tf.float32).
        minval (float): The minimum value for clipping. Defaults to 0.0.
        maxval (float): The maximum value for clipping.
        cmap (str): A valid name for a matplotlib color map or `gray`.
        cycle_color_map (int): Repeats the colormap the defined number of times in a
            continuous permutation pattern. Useful if high ranges within `tensor` shall
            be visualized in high color resolution. Defaults to 1.
        mask (tf.Tensor): A boolean mask with the same shape as `tensor` that indicates
            which indices shall be excluded from the color mapping.
            Tensor(shape=[H, W], dtype=tf.bool).
        exclude_color (tuple): An (R, G, B) color in range [0, 255] that defines the
            color of the indices that are excluded by `mask`. Defaults to black (0, 0, 0).
        expand_height (int): A factor that specifies how many multiples of each element
            in H are produced. Useful if H << W. Default to 1.
        scope (str): The name scope of the function.

    Returns:
        Tensor(shape=[1, H * expand_height, W, 3], dtype=tf.uint8): An RGB image.
            The first dimension corresponds to the `batch_size` to be published with `tf.summary.image`.

    Raises:
        ValueError: When shape of `tensor` and `mask` are not equal if `mask` is not None.
        ValueError: When type of `mask` is not tf.bool if `mask` is not None.
        ValueError: If `maxval` is lower or equal to `minval`.

    TODO:
        * add support for batched tensors
        * add support for expand_width
    """

    if mask is not None:
        if not tensor.shape == mask.shape:
            raise ValueError('`tensor` and `mask` must have same shape. {} vs. {}'.format(
                tensor.shape, mask.shape))
        if not mask.dtype == tf.bool:
            raise TypeError('`mask` must have type `tf.bool` but is {}'.format(mask.dtype))

    if maxval is not None and maxval <= minval:
        raise ValueError('Clipping values should be defined as `minval` < `maxval`. '
                         'Got {} !< {}.'.format(minval, maxval))

    if not isinstance(cycle_color_map, int) or cycle_color_map < 1:
        raise ValueError('`cycle_color_map` must be an integer greater than 1.')

    with tf.name_scope(name=scope, default_name='create_image'):

        # normalization
        maxval = maxval if maxval is not None else tf.reduce_max(tensor)
        tensor = tf.clip_by_value(tensor, minval, maxval)
        tensor = tf.round((tensor - minval) * 255 * cycle_color_map) / (maxval - minval)
        indices = tf.cast(tensor, dtype=tf.int32)

        # colorization
        if cmap == 'gray':
            colors = tf.range(start=0, limit=256, delta=1, dtype=tf.uint8)
            tf_image = tf.stack([tf.gather(colors, indices), ] * 3, axis=-1)
        else:
            cmap_colors = get_cmap(cmap).colors
            cm = [cmap_colors if i % 2 else cmap_colors[::-1] for i in range(cycle_color_map)]
            colors = tf.cast(tf.concat(cm, axis=0) * 255, dtype=tf.uint8)
            tf_image = tf.gather(colors, indices)

        if mask is not None:
            stacked_mask = tf.stack([mask] * 3, axis=-1)
            tf_image = tf.where(stacked_mask,
                                tf_image,
                                exclude_color * tf.ones_like(stacked_mask, dtype=tf.uint8))

        # expand the image height for better visualization in TensorBoard
        tf_image = expand_image_height(tf_image, expand_height)
        tf_image = tf_image[tf.newaxis, :, :, :]

    return tf_image



# -*- coding: utf-8 -*-

__author__ = """Larissa Triess"""
__email__ = 'mail@triess.eu'

from .confusion_matrix import confusion_matrix_image
from .colorful_tensors import expand_image_height, create_image

__all__ = [
    'confusion_matrix_image',
    'expand_image_height',
    'create_image',
]



# -*- coding: utf-8 -*-

__author__ = """Larissa Triess"""
__email__ = 'mail@triess.eu'

from .layers import get_padding_sizes, pad, cyclic_padding

__all__= [
    'get_padding_sizes',
    'pad',
    'cyclic_padding',
]



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='tf_helpers',
    version='0.0.2',
    author='Larissa Triess',
    author_email='mail@triess.eu',
    description='A collection of useful functions for TensorFlow for training and visualization.',
    url='https://github.com/ltriess/tf_helpers',
    license='MIT',
    packages=find_packages(exclude=['docs', 'tests']),
    python_requires='>=3.5',
    install_requires=[
        'numpy',
        'matplotlib',
    ],
    extras_require={
        'tf': ['tensorflow'],
        'tf_gpu': ['tensorflow-gpu'],
    },
)
